"""
解析introduction, 自动生成.md

- 示例:
```
标题
- 这一定是简介, 没有的话, 简介就是标题

GET /api/authors/InstitutionType/   # 接口, 可能会在最前方指明请求类型[GET, POST]
POST /api/authors/InstitutionType/      # 注明POST方法的话, 会将后面携带的dict加上
{
    "请求数据": "xxx"
}
```

"""


import os
import sys
from jinja2 import Template
from bdtime import tt

from django.conf import settings
from django.http import HttpResponse
from django.shortcuts import render
from django.core.cache import cache
from rest_framework.request import Request
from rest_framework.views import APIView

from .django import APIResponse, BaseListView, api_decorator
from .MultipleConditionSearch import AdvancedSearchView
from .django import convert_db_field_type_to_python_type
from .pure import convert_query_parameter_to_bool
from .django import get_base_serializer
from .django import my_api_assert_function


global tmp


RLT_PATH_OF_JINJIA2_TEMPLATE = 'templates\\autoWiki.html'       # 使用`jinja2`模板来生成`wiki`模板
ABS_PATH_OF_JINJIA2_TEMPLATE = os.path.join(os.path.dirname(__file__), RLT_PATH_OF_JINJIA2_TEMPLATE)


class AutoWiki(APIView):
    """
    自动生成wiki文档

    - app_name: 目标app, 必填
    - view_class_name: 目标view, 可空
    - ret_all_fields: 是否返回全部字段, 默认False
    - ADD_CAN_BE_EMPTY: 增加“能否为空”字段, 默认False
    - get_output_file: 下载文件, 仅能下一个*view_class*的说明文档, 适合*linux*服务端上使用

    - `.md`备注方法
        {#[*]( "这是注释")#}
        {#[ ]( "最好写上 [GET POST] 等")#}
    """

    open_output_txt_file = True     # `win`在运行完成后, 打开输出的`txt`文件
    path_of_jinja2_template = ABS_PATH_OF_JINJIA2_TEMPLATE       # 使用`jinja2`模板来生成`wiki`模板

    def get(self, request, *args, **kwargs):
        """自动生成wiki文档"""
        res_context_dc = {}     # 将结果变量储存为文本字典, 好用`jinja2`来填写模板

        ret = []
        request_data = request.GET

        # --- 必填
        app_name = request_data.get('app_name')
        assert app_name, 'app_name不能为空!'

        # --- 选填
        view_class_name = request_data.getlist('view_class_name')       # 可空
        ret_all_fields = request_data.getlist('ret_all_fields', False)     # 返回全部字段 or 只保留serializer_fields中返回的字段
        ADD_CAN_BE_EMPTY = request_data.getlist('ADD_CAN_BE_EMPTY', False)   # 增加“能否为空”字段
        SLEEP_TIME = request_data.get('SLEEP_TIME', 0.2)    # sleep时间, 也就是文件保存到打开的间隔
        MAX_LEN = request_data.get('MAX_LEN', 100)      # 示例数据最长字段限制

        exec(f'global tmp; from {app_name} import views as tmp;')        # 执行
        print('---------', tmp)

        views = tmp
        func_ls = dir(views)

        s = 0
        vs = []
        for f in func_ls:
            md = getattr(views, f)
            if hasattr(md, 'queryset') and hasattr(md, 'serializer_class'):
                print('---------', f, md)
                if getattr(md, 'queryset') is not None and (getattr(md, 'serializer_class') or getattr(md, 'list_serializer_class') or getattr(md, 'auto_generate_serializer_class') or getattr(md, 'retrieve_serializer_class')):
                    if not view_class_name or f in view_class_name:
                        s += 1
                        print(f, md)
                        vs.append([f, md])
        print(s)
        if s == 0:
            msg = '没有任何可以转换的view_class! 将自动生成!!'
            return APIResponse(None, status=404, msg=msg)

        print('--------------------------------------------------------------------------')

        """生成一个view的wiki文档"""
        if 1:
            """生成一个view的wiki文档"""
            from bddjango.django import get_base_model
            from django.forms import model_to_dict
            from bddjango import show_json
            from bddjango.django import get_field_type_in_py
            from bddjango import create_dir_if_not_exist, create_file_if_not_exist
            import re
            import json
            import os

            tempdir_rootpath = 'tempdir'        # 临时输出文件的根目录
            output_dirpath = os.path.join(tempdir_rootpath, 'output')       # 临时输出文件的子目录
            create_dir_if_not_exist(tempdir_rootpath)
            create_dir_if_not_exist(output_dirpath)

            path_of_jinja2_template = request_data.get('path_of_jinja2_template', self.path_of_jinja2_template)

            for view_id in range(len(vs)):
                # view_id = 0       # debug
                # issubclass(v, BaseListView)

                # --- 获取单个view的名称和实例
                fv = vs[view_id]
                f, v = fv       # view_class_name, view_class_instance

                # --- 输出文件的路径
                output_f_suffix = '.txt'        # 使用`.md`打开太慢, `typora`会报错
                # output_f_suffix = '.md' if path_of_jinja2_template else '.txt'

                output_fname = f + output_f_suffix
                output_fname = os.path.join(output_dirpath, output_fname)
                if os.path.exists(output_fname):
                    os.remove(output_fname)

                def output_to_file_and_prt(text='', output_fname=output_fname, show_in_console=False):

                    output_file = open(output_fname, 'a+', encoding='utf-8')

                    if show_in_console:
                        print(text)
                    print(text, file=output_file)
                    output_file.close()

                # --- 开始写入

                output_to_file_and_prt('****************')
                res_context_dc['introduction'] = v.__doc__      # context: 简介
                output_to_file_and_prt(res_context_dc['introduction'])

                output_to_file_and_prt('****************')
                output_to_file_and_prt()

                output_to_file_and_prt(f'\n=======  {view_id}  =========  \n')
                res_context_dc['view_id'] = str(view_id)  # context: view_id

                view_class_name, view_class_instance = f, v
                output_to_file_and_prt(f'---, view_class_name: [{view_class_name}], view_class_instance: [{view_class_instance}]')
                res_context_dc['view_class_name'] = str(view_class_name)     # context: view_class_name
                res_context_dc['view_class_instance'] = str(view_class_instance)     # context: view_class_instance

                output_to_file_and_prt()

                # --- 获取model和meta
                md = get_base_model(v.queryset)
                meta = md._meta

                url = f'/api/{meta.app_label}/{f}/'

                output_to_file_and_prt('------------ 请求URL')
                output_to_file_and_prt(f"`{url}`")
                res_context_dc['request_url'] = str(url)     # context: 请求URL

                output_to_file_and_prt()

                if hasattr(v, 'filter_fields'):
                    if v.filter_fields:
                        output_to_file_and_prt('------------ 过滤字段')
                        output_to_file_and_prt(f"`{v.filter_fields}`")
                        res_context_dc['filter_fields'] = str(v.filter_fields)  # context: 请求URL
                        output_to_file_and_prt()

                # 确定serializer_class, 以免报错
                v.serializer_class = v.serializer_class or v.retrieve_serializer_class or v.list_serializer_class

                # --- 获取field_name和verbose_name

                field_names = [field.name for field in meta.fields]
                verbose_names = [field.verbose_name for field in meta.fields]
                can_be_empty_ls = [field.null and field.blank for field in meta.fields]

                field_type_ls = [get_field_type_in_py(md, field_name) for field_name in field_names]
                # field_type = get_field_type_in_py(md, field_name)

                if hasattr(v, 'auto_generate_serializer_class') and v.auto_generate_serializer_class:
                    # from bddjango import get_base_serializer
                    # v: BaseListView
                    v__serializer_class = v().get_serializer_class()
                    # v__queryset = get_base_model(v().get_queryset()).objects.all()
                    # v__serializer_class(v__queryset, many=True).data
                    # v__serializer_class.__dict__.get('_declared_fields')
                    v.serializer_class = v__serializer_class

                # --- 把serializers里面的拓展字段通过__doc__转义后加进field_names
                serializer_field_ls = v.serializer_class.__dict__.get('_declared_fields')
                text = v.serializer_class.__doc__ or ""

                for serializer_field_i, field_type_i in serializer_field_ls.items():
                    reg = re.compile(f'^.*{serializer_field_i}: +(.*?) *$', re.M)
                    match = reg.search(text)
                    if match:
                        verbose_name_i = match.group(1)
                    else:
                        verbose_name_i = serializer_field_i

                    field_type_i = convert_db_field_type_to_python_type(str(field_type_i).replace('()', ''))
                    print(serializer_field_i, verbose_name_i, field_type_i)
                    can_be_empty_i = True

                    field_names.append(serializer_field_i)
                    verbose_names.append(verbose_name_i)
                    field_type_ls.append(field_type_i)
                    can_be_empty_ls.append(can_be_empty_i)

                # --- 获取serializer_class
                serializer_class = v.serializer_class

                if issubclass(v, BaseListView):
                    if serializer_class is None:
                        if v.list_serializer_class:
                            serializer_class = v.list_serializer_class
                        if v.retrieve_serializer_class:
                            # 有限详情页的序列化器
                            serializer_class = v.retrieve_serializer_class

                # --- 示例字段
                smeta = serializer_class.Meta
                if hasattr(smeta, 'fields'):
                    sf = serializer_class.Meta.fields
                    if ret_all_fields is True or sf == '__all__':
                        # sf = '__all__'
                        sf = serializer_class.Meta.fields = field_names
                else:
                    sf = field_names.copy()
                    exclude_ls = smeta.exclude
                    if exclude_ls:
                        for e in exclude_ls:
                            sf.remove(e)

                try:
                    # 直接调用view的get接口
                    result = v.as_view()(request._request).data.get('result')

                    if isinstance(result, dict):
                        dc_ls = result.get('data')
                    elif isinstance(result, list):
                        dc_ls = result
                    else:
                        raise TypeError(f'result为未知的返回类型! {type(result)}')
                    dc_ls = dc_ls[:3]
                    if not dc_ls:
                        raise ValueError('\n============== 返回的样例数据`example_data`为空! =============\n')

                except Exception as e:
                    try:
                        print('--- Error! 调用view方法失败, 尝试使用model方案... ', e)
                        # --- 示例数据, 用model
                        qs_ls = md.objects.all()[:3]
                        # q0 = qs_ls[0]     # 单个数据
                        try:
                            dc_ls = serializer_class(qs_ls, many=True).data
                        except Exception as e:
                            print('--- Error! 序列化出错! 可能未携带annotate所需字段? 将自动生成基本序列化器! --- ' + str(e))
                            serializer_class = get_base_serializer(qs_ls)
                            dc_ls = serializer_class(qs_ls, many=True).data

                        dc_ls = [dict(dc) for dc in dc_ls]      # 返回数据示例
                    except Exception as e:
                        print(f'========== 错误信息: view_class_name[{view_class_name}], view_class_instance[{v}]')
                        raise e

                # -- 简化过长的字段
                for dc in dc_ls:
                    for k, v in dc.items():
                        if isinstance(v, str):
                            if len(v) > MAX_LEN:
                                dc[k] = v[:MAX_LEN] + '...略'
                    print(dc)

                output_to_file_and_prt('---------- 示例数据')
                example_data = json.dumps(dc_ls, sort_keys=False, indent=4, separators=(', ', ': '), ensure_ascii=False)
                output_to_file_and_prt(example_data)
                res_context_dc['example_data'] = str(example_data)  # context: 示例数据
                output_to_file_and_prt()

                # --- 参数说明
                print(field_names, verbose_names)

                if ADD_CAN_BE_EMPTY:
                    ss = "| 类型 | 字段名 | 说明 | 能否为空 |\n| --- | --- | --- | --- |\n"
                else:
                    ss = "| 类型 | 字段名 | 说明 |\n| --- | --- | --- |\n"

                for field_name, verbose_name, can_be_empty, field_type in zip(field_names, verbose_names, can_be_empty_ls, field_type_ls):
                    print(field_name, verbose_name, can_be_empty)

                    si = ''
                    if sf == '__all__' or field_name in sf:
                        # field_type = get_field_type_in_py(md, field_name)
                        if ADD_CAN_BE_EMPTY:
                            si = f'| {field_type} | {field_name} | {verbose_name} | {can_be_empty} |\n'
                        else:
                            si = f'| {field_type} | {field_name} | {verbose_name} |\n'
                    ss += si

                output_to_file_and_prt('---------- 参数说明')
                parameters_explain = ss
                output_to_file_and_prt(parameters_explain)
                res_context_dc['parameters_explain'] = str(parameters_explain)  # context: 参数说明
                output_to_file_and_prt()

                open_output_txt_file = request_data.get('open_output_txt_file', self.open_output_txt_file)

                if convert_query_parameter_to_bool(open_output_txt_file):
                    if convert_query_parameter_to_bool(path_of_jinja2_template):
                        # abs_path = os.path.join(settings.BASE_DIR, 'authors/myTestTemplates.html')
                        # path_of_jinja2_template = path_of_jinja2_template if os.path.exists(path_of_jinja2_template) else os.path.exists(os.path.join(settings.BASE_DIR, path_of_jinja2_template))
                        path_of_jinja2_template = os.path.join(settings.BASE_DIR, path_of_jinja2_template)
                        path_of_jinja2_template = path_of_jinja2_template.replace('/', '\\')
                        my_api_assert_function(os.path.exists(path_of_jinja2_template), f'path_of_jinja2_template路径不存在![{path_of_jinja2_template}]')
                        print('--- res_context_dc:', res_context_dc)

                        # region # --- 解析`introduction`, 获取[标题, 简要描述, 请求url]等字段
                        introduction = res_context_dc.get('introduction')
                        if introduction:
                            # --- 获取标题`introduction_title`
                            reg = re.compile(r'^\n(.+?)\n')
                            match = reg.search(introduction)
                            if match:
                                introduction_title = match.group(1).strip()
                                introduction_title = re.sub(r'^#+ +', '', introduction_title)   # 去掉开头的`#`
                                res_context_dc['introduction_title'] = introduction_title

                            # --- 获取示例url`introduction_url`
                            # introduction = """
                            #     学者成果高级检索
                            #     - 必须指定author_code, 因为是专家详情页
                            #     - 使用post方法进行复合检索
                            #     - 检索字段名为`Q_add_ls`
                            #
                            #     POST /api/authors/ThesisAdvancedSearch/
                            #     {
                            #         "order_type_ls": ["title", "id"],
                            #         "distinct_field_ls": ["title"],
                            #         "page_size": 3,
                            #         "Q_add_ls": [
                            #             {
                            #                 "add_logic": "and",
                            #                 "Q_ls": [
                            #                     {
                            #                         "add_logic": "and",
                            #                         "search_field": "author_code",
                            #                         "search_keywords": "29856",
                            #                         "accuracy": "1"
                            #                     },
                            #                     {
                            #                         "add_logic": "and",
                            #                         "search_field": "title",
                            #                         "search_keywords": "可持续",
                            #                         "accuracy": "0"
                            #                     }
                            #                 ]
                            #             },
                            #             {
                            #                 "add_logic": "and",
                            #                 "Q_ls": [
                            #                     {
                            #                         "add_logic": "and",
                            #                         "search_field": "published_date",
                            #                         "search_keywords": "2019-02-10",
                            #                         "accuracy": "gte"
                            #                     },
                            #                     {
                            #                         "add_logic": "and",
                            #                         "search_field": "published_date",
                            #                         "search_keywords": "2019-04-01",
                            #                         "accuracy": "lte"
                            #                     }
                            #                 ]
                            #             }
                            #         ]
                            #     }
                            #     """
                            reg = re.compile(r'.*?/.+/.+/.*')
                            match = reg.findall(introduction)
                            if match:
                                _introduction_url = []
                                for m_i in match:
                                    m_i = m_i.strip()
                                    if 'POST' in m_i:
                                        reg = re.compile(m_i + '.*?' + r'(\{.*\}).*$', re.S)      # 只能匹配POST后的一个大括号!
                                        findall_ls = reg.findall(introduction)
                                        if findall_ls:
                                            findall_i = findall_ls[0]
                                            try:
                                                findall_str = json.dumps(json.loads(findall_i), sort_keys=False, indent=4, separators=(', ', ': '), ensure_ascii=False)
                                            except Exception as e:
                                                e_str = '****** POST请求携带的字典错误!' + str(e)
                                                raise ValueError(e_str)
                                            m_i += '\n' + findall_str
                                    _introduction_url.append(m_i)
                                introduction_url = '\n'.join(_introduction_url)
                                res_context_dc['introduction_url'] = introduction_url

                            reg = re.compile(r'\n( *?- .+)')
                            match = reg.findall(introduction)
                            # from bddjango import show_ls
                            # show_ls(match)
                            if match:
                                _introduction_summary = []
                                blank_ls = []
                                for summary_i in match:
                                    summary_i: str
                                    m1 = re.match(r'^ *', summary_i)
                                    blank_i = m1.span()[-1] if m1 else 0
                                    blank_ls.append(blank_i)
                                    _introduction_summary.append(summary_i)
                                blank_min = min(blank_ls)       # 将所有简介缩进到最左侧, 不然会变成代码块!
                                _introduction_summary = [s_i[blank_min:] for s_i in _introduction_summary]
                                introduction_summary = '\n'.join(_introduction_summary)
                            else:
                                introduction_summary = '- ' + introduction_title
                            res_context_dc['introduction_summary'] = introduction_summary
                        # endregion

                        # region # --- 分析属于那种请求类型`context_request_type`
                        view_class_type = 'APIView'
                        if hasattr(view_class_instance, '_name'):
                            view_class_type = getattr(view_class_instance, '_name')

                        context_request_type = '其它'
                        if view_class_type == 'BaseListView':
                            context_request_type = '基本查找类'
                        elif view_class_type == 'CompleteModelView':
                            context_request_type = '增删查改类'
                        elif view_class_type == 'AdvancedSearchView':
                            context_request_type = '高级检索类'
                        res_context_dc['context_request_type'] = context_request_type
                        # endregion

                        # region # --- 开始填充jinja模板
                        with open(path_of_jinja2_template, encoding='utf-8') as file_:
                            template = Template(file_.read())

                        # from jinja2 import PackageLoader, Environment, FileSystemLoader
                        # env = Environment(loader=PackageLoader('python_project', 'templates'))  # 创建一个包加载器对象

                        # env = Environment(loader=FileSystemLoader(app_name))  # 文件加载器, 可用`list_templates`方法查看存在哪些东西
                        # os.path.exists(path_of_jinja2_template)
                        # template = env.get_template(path_of_jinja2_template)
                        content = template.render(**res_context_dc)
                        with open(output_fname, 'w', encoding='utf-8') as f:
                            f.write(content)
                        # endregion

                    # 获取output文件, 仅能获取一个view_class
                    get_output_file = convert_query_parameter_to_bool(request_data.get('get_output_file'))
                    if get_output_file:
                        newFileName = output_fname
                        with open(output_fname, 'rb') as f:
                            response = HttpResponse(f.read(), content_type='application/txt')
                            response['Content-Disposition'] = 'attachment; filename="{}"'.format(newFileName)
                            response["Access-Control-Allow-Origin"] = '*'
                            response["Server"] = '*'
                            return response
                    else:
                        if sys.platform == 'win32':
                            for i in range(5):
                                tt.sleep(1)
                                print('111')
                                if os.path.exists(output_fname):
                                    print('222')
                                    os.startfile(output_fname)
                                    tt.sleep(SLEEP_TIME)
                                    # tt.sleep(10)
                                    # os.remove(output_fname)
                                    from bddjango.adminclass import remove_temp_file
                                    remove_temp_file('tempdir', MAX_TEMPS=10)
                                    break
                        else:
                            print('--- output_fname: ', output_fname)
                            with open(output_fname, 'r') as f:
                                print(f.read())

                                ret.append({output_fname: f.read()})

        ret_dc = {
            'res_context_dc': res_context_dc
        }
        ret.append(ret_dc)

        # output_file.close()
        return APIResponse(ret, 200, 'ok')


from warnings import warn
from .pure import *


def version():
    """
    * 2021/3/14
    - [pypi_url](https://pypi.org/project/bddjango/)

    # 党史项目结束, 海南项目开始

    ## 2.1修改部分
        - 修复了AdvancedSearchView post时page_size失效的问题     # 2.1.1
        - 对`list`的数据返回格式进行了优化, 避免了自定义page_size时无效的bug
        - `BaseListView`增加`set_request_data`方法
        - 优化`distince_type_ls`和`order_type_ls`的错误提示
        - `AdvancedSearchView`增加`Q_add_ls`检索方法      # 2.1.2
        - `.pure`增加`convert_query_parameter_to_bool`函数
        - '.django'增加`update_none_to_zero_by_field_name`函数
        - 增加`AliasField`绑定字段方法      # 2.1.3
        - 按年份分布的统计方法`get_df_by_freq_and_year`       # 2.1.4
        - 修复distinct_field_ls和order_type_ls的冲突      # 2.1.5
        - BaseListView中, 加入'pk'作为默认过滤字段     # 2.1.6
        - template调整`simpleui_change_list_actions.html`样式
        - 新增`only_get_distinct_field`功能
        - 修复2.1.6版本忘记上传templates导致adminclass导入失败问题      # 2.1.7
        - 修复了`BaseListView`的list方法中`get_serializer_context`无效的问题, 在`paginate_qsls_to_dcls`加入了`context`参数       # 2.1.8
        - 在`adminclass.remove_temp_file`方法中增加描述字段`desc`, 并增加部分debug说明文字
        - 前端可以控制`base_fields`参数, 但仅适用于`auto_generate_serializer`为`True`的情况
        - 加上`auth.MyValidationError`, 修复验证模块报错bug
        - 简化部分代码: 将(获取self.key or 获取query_dc.key)的方法统一  # 2.1.9
        - 修复`_get_key_from_query_dc_or_self`的关键bug!     # 2.1.10
        - 拓展`get_MySubQuery`方法, 对`annotate`生成的字段生效      # 2.1.11
        - 在`get_abs_order_type_ls`方法中处理str类型数据
        - 自动生成wiki打包autoWiki.py        # 2.1.12

    # 2.2
        - 方法version转移到__init__.py中
        - 修改由`get_key_from_query_dc_or_self`引起的bug
        - 高级检索类增加`search_ls_dc`和`search_conf`两个参数
        - 修改由于readme.md出错导致pypi上传失败的bug	# 2.2.2
        - 去掉部分print	# 2.2.3
    """
    v = "2.2.3"     # 正式版: 2.2.3
    return v


try:
    from .django import *
except Exception as e:
    warn('导入django失败? --- ' + str(e))

try:
    from .myFIelds import AliasField        # 这个只能在这里引用, 不然`adminclass`报错
except Exception as e:
    warn('导入`AliasField`失败? --- ' + str(e))




"""
纯python的功能函数
"""
import json
import os
import pandas as pd


def add_status_and_msg(dc_ls, status=200, msg=None):
    if status != 200 and msg is None:
        msg = '请求数据失败!'

    if status == 200 and msg is None:
        msg = "ok"

    ret = {
        'status': status,
        'msg': msg,
        'result': dc_ls
    }
    return ret


def show_json(data: dict, sort_keys=False):
    try:
        print(json.dumps(data, sort_keys=sort_keys, indent=4, separators=(', ', ': '), ensure_ascii=False))
    except:
        if isinstance(data, dict):
            for k, v in data.items():
                print(k, ' --- ', v)
        else:
            for k, v in data:
                print(k, ' --- ', v)


def show_ls(data: list, ks=None):
    for dc in data:
        if ks:
            if isinstance(ks, str):
                ks = [ks]
            d = [dc.get(k) for k in ks]
        else:
            d = dc
        print(d)


def add_space_prefix(text, n, more=True, prefix='\u3000'):
    text = str(text)
    if more:
        ret = prefix * n + text
    else:
        ret = prefix * (n - len(text)) + text
    return ret


def create_file_if_not_exist(file_name):
    if not os.path.exists(file_name):
        with open(file_name, 'w') as f:
            f.write('')
        return False
    return True


def create_dir_if_not_exist(dirpath: str):
    if not os.path.exists(dirpath):
        os.mkdir(dirpath)
        return False
    return True


import inspect
import functools


def get_class_that_defined_method(meth):
    """
    get mehod's class
    """
    if isinstance(meth, functools.partial):
        return get_class_that_defined_method(meth.func)
    if inspect.ismethod(meth) or (inspect.isbuiltin(meth) and getattr(meth, '__self__', None) is not None and getattr(meth.__self__, '__class__', None)):
        for cls in inspect.getmro(meth.__self__.__class__):
            if meth.__name__ in cls.__dict__:
                return cls
        meth = getattr(meth, '__func__', meth)  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
        cls = getattr(inspect.getmodule(meth),
                      meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0],
                      None)
        if isinstance(cls, type):
            return cls
    return getattr(meth, '__objclass__', None)  # handle special descriptor objects


def get_whole_codename_by_obj_and_perm(obj=None, perm=None, suffix_model_name=False):
    """
    得到obj的perm对应的完整codename: whole_codename

    :param obj: 模型 or 对象
    :param perm: 权限名
    :param suffix_model_name: perm里边没有obj对应model的model_name, 需要函数手动添加
    :return:
    """
    if obj:
        if suffix_model_name:
            ret = f'{obj._meta.app_label}.{perm}_{obj._meta.model_name}'
        else:
            ret = f'{obj._meta.app_label}.{perm}'
    else:
        ret = perm
    return ret


def conv_df_to_serializer_data(df) -> list:
    assert isinstance(df, pd.DataFrame), 'df的类型必须是DataFrame!'
    ret_ls = []
    for index, row in df.iterrows():
        k = row.index.tolist()
        v = row.values.tolist()
        data = dict(zip(k, v))

        ret_ls.append(data)
    return ret_ls


def convert_query_parameter_to_bool(query_parameter, false_ls=None):
    """
    将请求参数转化为`bool`类型

    :param query_parameter: 请求参数
    :param false_ls: 将转换为`false`的值
    :return: bool, true or false
    """
    if not false_ls:
        false_ls = ['0', 0, None, 'None', 'Null', [], {}, 'False', 'false', '', 'null']
    ret = query_parameter not in false_ls
    return ret




from django.shortcuts import render
from . import models, serializers
from rest_framework.views import APIView
from django.core.cache import cache
from rest_framework.request import Request
import sys

from bddjango.django import APIResponse, BaseListView, api_decorator
from bddjango.MultipleConditionSearch import AdvancedSearchView
from bddjango import convert_db_field_type_to_python_type
from bddjango import convert_query_parameter_to_bool
from bdtime import tt

global tmp


class MyTest(APIView):
    """
    自动生成wiki文档

    - app_name: 目标app, 必填
    - view_class_name: 目标view, 可空
    - ret_all_fields: 是否返回全部字段, 默认False
    - ADD_CAN_BE_EMPTY: 增加“能否为空”字段, 默认False



    {#[*]( "这是注释")#}

    {#[ ]( "最好写上 [GET POST] 等")#}
    """

    open_output_txt_file = True  # `win`在运行完成后, 打开输出的`txt`文件
    path_of_jinja2_template = 'myTestTemplates.html'  # 使用`jinja2`模板来生成`wiki`模板

    def get(self, request, *args, **kwargs):
        """自动生成wiki文档"""
        res_context_dc = {}  # 将结果变量储存为文本字典, 好用`jinja2`来填写模板

        ret = []
        request_data = request.GET

        # --- 必填
        app_name = request_data.get('app_name')
        assert app_name, 'app_name不能为空!'

        # --- 选填
        view_class_name = request_data.getlist('view_class_name')  # 可空
        ret_all_fields = request_data.getlist('ret_all_fields', False)  # 返回全部字段 or 只保留serializer_fields中返回的字段
        ADD_CAN_BE_EMPTY = request_data.getlist('ADD_CAN_BE_EMPTY', False)  # 增加“能否为空”字段
        SLEEP_TIME = request_data.get('SLEEP_TIME', 0.2)  # sleep时间, 也就是文件保存到打开的间隔
        MAX_LEN = request_data.get('MAX_LEN', 100)  # 示例数据最长字段限制

        exec(f'global tmp; from {app_name} import views as tmp;')  # 执行
        print('---------', tmp)

        views = tmp
        func_ls = dir(views)

        s = 0
        vs = []
        for f in func_ls:
            md = getattr(views, f)
            if hasattr(md, 'queryset') and hasattr(md, 'serializer_class'):
                print('---------', f, md)
                if getattr(md, 'queryset') is not None and (
                        getattr(md, 'serializer_class') or getattr(md, 'list_serializer_class') or getattr(md,
                                                                                                           'auto_generate_serializer_class') or getattr(
                        md, 'retrieve_serializer_class')):
                    if not view_class_name or f in view_class_name:
                        s += 1
                        print(f, md)
                        vs.append([f, md])
        print(s)
        if s == 0:
            msg = '没有任何可以转换的view_class! 将自动生成!!'
            return APIResponse(None, status=404, msg=msg)

        print('--------------------------------------------------------------------------')

        """生成一个view的wiki文档"""
        if 1:
            """生成一个view的wiki文档"""
            from bddjango.django import get_base_model
            from django.forms import model_to_dict
            from bddjango import show_json
            from bddjango.django import get_field_type_in_py
            from bddjango import create_dir_if_not_exist, create_file_if_not_exist
            import re
            import json
            import os

            tempdir_rootpath = 'tempdir'  # 临时输出文件的根目录
            output_dirpath = os.path.join(tempdir_rootpath, 'output')  # 临时输出文件的子目录
            create_dir_if_not_exist(tempdir_rootpath)
            create_dir_if_not_exist(output_dirpath)

            path_of_jinja2_template = request_data.get('path_of_jinja2_template', self.path_of_jinja2_template)

            for view_id in range(len(vs)):
                # view_id = 0       # debug
                # issubclass(v, BaseListView)

                # --- 获取单个view的名称和实例
                fv = vs[view_id]
                f, v = fv  # view_class_name, view_class_instance

                # --- 输出文件的路径
                output_f_suffix = '.txt'  # 使用`.md`打开太慢, `typora`会报错
                # output_f_suffix = '.md' if path_of_jinja2_template else '.txt'

                output_fname = f + output_f_suffix
                output_fname = os.path.join(output_dirpath, output_fname)
                if os.path.exists(output_fname):
                    os.remove(output_fname)

                def output_to_file_and_prt(text='', output_fname=output_fname, show_in_console=False):

                    output_file = open(output_fname, 'a+', encoding='utf-8')

                    if show_in_console:
                        print(text)
                    print(text, file=output_file)
                    output_file.close()

                # --- 开始写入

                output_to_file_and_prt('****************')
                res_context_dc['introduction'] = v.__doc__  # context: 简介
                output_to_file_and_prt(res_context_dc['introduction'])

                output_to_file_and_prt('****************')
                output_to_file_and_prt()

                output_to_file_and_prt(f'\n=======  {view_id}  =========  \n')
                res_context_dc['view_id'] = str(view_id)  # context: view_id

                view_class_name, view_class_instance = f, v
                output_to_file_and_prt(
                    f'---, view_class_name: [{view_class_name}], view_class_instance: [{view_class_instance}]')
                res_context_dc['view_class_name'] = str(view_class_name)  # context: view_class_name
                res_context_dc['view_class_instance'] = str(view_class_instance)  # context: view_class_instance

                output_to_file_and_prt()

                # --- 获取model和meta
                md = get_base_model(v.queryset)
                meta = md._meta

                url = f'/api/{meta.app_label}/{f}/'

                output_to_file_and_prt('------------ 请求URL')
                output_to_file_and_prt(f"`{url}`")
                res_context_dc['request_url'] = str(url)  # context: 请求URL

                output_to_file_and_prt()

                if hasattr(v, 'filter_fields'):
                    if v.filter_fields:
                        output_to_file_and_prt('------------ 过滤字段')
                        output_to_file_and_prt(f"`{v.filter_fields}`")
                        res_context_dc['filter_fields'] = str(v.filter_fields)  # context: 请求URL
                        output_to_file_and_prt()

                # 确定serializer_class, 以免报错
                v.serializer_class = v.serializer_class or v.retrieve_serializer_class or v.list_serializer_class

                # --- 获取field_name和verbose_name

                field_names = [field.name for field in meta.fields]
                verbose_names = [field.verbose_name for field in meta.fields]
                can_be_empty_ls = [field.null and field.blank for field in meta.fields]

                field_type_ls = [get_field_type_in_py(md, field_name) for field_name in field_names]
                # field_type = get_field_type_in_py(md, field_name)

                if v.auto_generate_serializer_class:
                    from bddjango import get_base_serializer
                    v: BaseListView
                    v__serializer_class = v().get_serializer_class()
                    v__queryset = get_base_model(v().get_queryset()).objects.all()
                    v__serializer_class(v__queryset, many=True).data
                    v__serializer_class.__dict__.get('_declared_fields')
                    v.serializer_class = v__serializer_class

                # --- 把serializers里面的拓展字段加进field_names
                serializer_field_ls = v.serializer_class.__dict__.get('_declared_fields')
                text = v.serializer_class.__doc__ or ""

                for serializer_field_i, field_type_i in serializer_field_ls.items():
                    reg = re.compile(f'^.*{serializer_field_i}: +(.*?) *$', re.M)
                    match = reg.search(text)
                    if match:
                        verbose_name_i = match.group(1)
                    else:
                        verbose_name_i = serializer_field_i

                    field_type_i = convert_db_field_type_to_python_type(str(field_type_i).replace('()', ''))
                    print(serializer_field_i, verbose_name_i, field_type_i)
                    can_be_empty_i = True

                    field_names.append(serializer_field_i)
                    verbose_names.append(verbose_name_i)
                    field_type_ls.append(field_type_i)
                    can_be_empty_ls.append(can_be_empty_i)

                # --- 获取serializer_class
                serializer_class = v.serializer_class

                if issubclass(v, BaseListView):
                    if serializer_class is None:
                        if v.list_serializer_class:
                            serializer_class = v.list_serializer_class
                        if v.retrieve_serializer_class:
                            # 有限详情页的序列化器
                            serializer_class = v.retrieve_serializer_class

                # --- 示例字段
                smeta = serializer_class.Meta
                if hasattr(smeta, 'fields'):
                    sf = serializer_class.Meta.fields
                    if ret_all_fields is True or sf == '__all__':
                        # sf = '__all__'
                        sf = serializer_class.Meta.fields = field_names
                else:
                    sf = field_names.copy()
                    exclude_ls = smeta.exclude
                    if exclude_ls:
                        for e in exclude_ls:
                            sf.remove(e)

                if 0:
                    # 这里尝试用drf的测试方法去获取返回示例数据
                    from rest_framework.test import APIRequestFactory
                    view = v.as_view()
                    print(view)
                    v_str = str(v)
                    example_url = v.__doc__

                    reg = re.compile(r"\.(\w+)\'>$")
                    match = reg.search(v_str)
                    v_name = match.group(1)
                    v_url = f'api/{app_name}/{v_name}'
                    # 找到url, 然后用测试模块测试
                    factory = APIRequestFactory()
                    rq = factory.get(v_url)
                    v.as_view()(rq).data
                    1

                    from flow_statistics.utils import get_statistics_result
                    get_statistics_result(v.queryset, get_serializer_data=False)
                    print('---', v.as_view()(request._request, *args, **kwargs).data)
                    v().get_list_ret(request._request, *args, **kwargs)

                    # v(request._request).as_view().get_queryset()
                    v.as_view()(request._request, *args, **kwargs).data

                    v.as_view()(request._request, *args, **kwargs)
                    v.as_view()(request._request, *args, **kwargs).data

                    v.as_view()
                    v(request).get_queryset()
                    type(v)
                    from flow_statistics.views import Statistics
                    get_statistics_result(v.queryset, get_serializer_data=False)

                    type(Statistics)
                    Statistics.get_queryset(Statistics)

                try:
                    # 直接调用view的get接口
                    result = v.as_view()(request._request).data.get('result')
                    if isinstance(result, dict):
                        dc_ls = result.get('data')
                    elif isinstance(result, list):
                        dc_ls = result
                    else:
                        raise TypeError(f'result为未知的返回类型! {type(result)}')
                    dc_ls = dc_ls[:3]

                except Exception as e:
                    print('--- Error! 调用view方法失败, 尝试使用model方案... ', e)
                    # --- 示例数据, 用model
                    qs_ls = md.objects.all()[:3]
                    # q0 = qs_ls[0]     # 单个数据
                    dc_ls = serializer_class(qs_ls, many=True).data
                    dc_ls = [dict(dc) for dc in dc_ls]  # 返回数据示例

                # -- 简化过长的字段
                for dc in dc_ls:
                    for k, v in dc.items():
                        if isinstance(v, str):
                            if len(v) > MAX_LEN:
                                dc[k] = v[:MAX_LEN] + '...略'
                    print(dc)

                output_to_file_and_prt('---------- 示例数据')
                example_data = json.dumps(dc_ls, sort_keys=False, indent=4, separators=(', ', ': '), ensure_ascii=False)
                output_to_file_and_prt(example_data)
                res_context_dc['example_data'] = str(example_data)  # context: 示例数据
                output_to_file_and_prt()

                # --- 参数说明
                print(field_names, verbose_names)

                if ADD_CAN_BE_EMPTY:
                    ss = "| 类型 | 字段名 | 说明 | 能否为空 |\n| --- | --- | --- | --- |\n"
                else:
                    ss = "| 类型 | 字段名 | 说明 |\n| --- | --- | --- |\n"

                for field_name, verbose_name, can_be_empty, field_type in zip(field_names, verbose_names,
                                                                              can_be_empty_ls, field_type_ls):
                    print(field_name, verbose_name, can_be_empty)

                    si = ''
                    if sf == '__all__' or field_name in sf:
                        # field_type = get_field_type_in_py(md, field_name)
                        if ADD_CAN_BE_EMPTY:
                            si = f'| {field_type} | {field_name} | {verbose_name} | {can_be_empty} |\n'
                        else:
                            si = f'| {field_type} | {field_name} | {verbose_name} |\n'
                    ss += si

                output_to_file_and_prt('---------- 参数说明')
                parameters_explain = ss
                output_to_file_and_prt(parameters_explain)
                res_context_dc['parameters_explain'] = str(parameters_explain)  # context: 参数说明
                output_to_file_and_prt()

                open_output_txt_file = request_data.get('open_output_txt_file', self.open_output_txt_file)

                if convert_query_parameter_to_bool(open_output_txt_file):
                    if convert_query_parameter_to_bool(path_of_jinja2_template):
                        """
                        解析introduction

                        `
                        标题
                        - 这一定是简介, 没有的话, 简介就是标题

                        /api/authors/InstitutionType/   # 接口, 可能会在最前方指明请求类型[GET, POST]
                        `

                        """
                        print('--- res_context_dc:', res_context_dc)

                        # region # --- 解析`introduction`, 获取[标题, 简要描述, 请求url]等字段
                        introduction = res_context_dc.get('introduction')
                        if introduction:

                            # --- 获取标题`introduction_title`
                            reg = re.compile(r'^\n(.+?)\n')
                            match = reg.search(introduction)
                            if match:
                                introduction_title = match.group(1).strip()
                                res_context_dc['introduction_title'] = introduction_title

                            # --- 获取示例url`introduction_url`
                            reg = re.compile(r'.*?/.+/.+/.*')
                            match = reg.findall(introduction)
                            if match:
                                _introduction_url = []
                                for m_i in match:
                                    m_i = m_i.strip()
                                    _introduction_url.append(m_i)
                                introduction_url = '\n'.join(_introduction_url)
                                res_context_dc['introduction_url'] = introduction_url

                            reg = re.compile(r'\n( *?- .+)')
                            match = reg.findall(introduction)
                            from bddjango import show_ls
                            show_ls(match)
                            if match:
                                _introduction_summary = []
                                for summary_i in match:
                                    _introduction_summary.append(summary_i)
                                introduction_summary = '\n'.join(_introduction_summary)
                            else:
                                introduction_summary = '- ' + introduction_title
                            res_context_dc['introduction_summary'] = introduction_summary
                        # endregion

                        # region # --- 分析属于那种请求类型`context_request_type`
                        view_class_type = 'APIView'
                        if hasattr(view_class_instance, '_name'):
                            view_class_type = getattr(view_class_instance, '_name')

                        context_request_type = '其它'
                        if view_class_type == 'BaseListView':
                            context_request_type = '基本查找类'
                        elif view_class_type == 'CompleteModelView':
                            context_request_type = '增删查改类'
                        elif view_class_type == 'AdvancedSearchView':
                            context_request_type = '高级检索类'
                        res_context_dc['context_request_type'] = context_request_type
                        # endregion

                        # region # --- 开始填充jinja模板
                        from jinja2 import PackageLoader, Environment, FileSystemLoader
                        # env = Environment(loader=PackageLoader('python_project', 'templates'))  # 创建一个包加载器对象
                        env = Environment(loader=FileSystemLoader(app_name))  # 文件加载器, 可用`list_templates`方法查看存在哪些东西
                        template = env.get_template(path_of_jinja2_template)
                        content = template.render(**res_context_dc)
                        with open(output_fname, 'w') as f:
                            f.write(content)
                        # endregion

                    if sys.platform == 'win32':
                        os.startfile(output_fname)
                        tt.sleep(SLEEP_TIME)
                        os.remove(output_fname)
                    else:
                        print('--- output_fname: ', output_fname)
                        with open(output_fname, 'r') as f:
                            print(f.read())

                            ret.append({output_fname: f.read()})

        ret_dc = {
            'res_context_dc': res_context_dc
        }
        ret.append(ret_dc)

        # output_file.close()
        return APIResponse(ret, 200, 'ok')



import os
import shutil
from django.conf import settings


TEMPDIR = 'tempdir'     # 临时文件夹


# --- 检测是否安装了simpleui, 以使用不同界面配置
app_ls: list = settings.INSTALLED_APPS
BD_USE_SIMPLEUI = True if 'simpleui' in app_ls or 'simpleuipro'in app_ls else False
CHANGE_LIST_HTML_PATH = os.path.join('entities', 'simpleui_change_list.html') if BD_USE_SIMPLEUI else os.path.join('entities', 'base_change_list.html')


# --- 默认检测是否simpleui + guadian, 配置object权限按钮的样式. ps: 优先使用settings中的设置.
use_guardian_settings_name = 'BD_USE_GUARDIAN'
CHANGE_FORM_TEMPLATE = 'admin/guardian/model/change_form.html'      # guardian的默认样式
if not hasattr(settings, use_guardian_settings_name):
    BD_USE_GUARDIAN = False
    if 'guardian' in app_ls:
        BD_USE_GUARDIAN = True
        if BD_USE_SIMPLEUI:
            CHANGE_FORM_TEMPLATE = 'admin/guardian/model/simpleui_guardian_change_form.html'
else:
    BD_USE_GUARDIAN = getattr(settings, use_guardian_settings_name)


# --- 加入django的templates路径
BASE_DIR = os.path.dirname(os.path.dirname(__file__))
TEMPLATES_DIR = os.path.join(BASE_DIR, 'templates')
settings.TEMPLATES[0]['DIRS'].append(TEMPLATES_DIR)

CHANGE_LIST_HTML_PATH = os.path.join(TEMPLATES_DIR, CHANGE_LIST_HTML_PATH)

# --- 如果使用的是simpleui, 则需要修复一些bug
# https://github.com/newpanjing/simpleui/issues/405
if BD_USE_SIMPLEUI and not os.path.exists(os.path.join('templates', 'admin', 'actions.html')):
    from django.contrib.admin.templatetags.admin_list import admin_actions, InclusionAdminNode, register
    template_name = 'simpleui_admin_actions.html'
    @register.tag(name='simpleui_admin_actions')
    def simpleui_admin_actions(parser, token):
        return InclusionAdminNode(parser, token, func=admin_actions, template_name=template_name)

    CHANGE_LIST_HTML_PATH = os.path.join(TEMPLATES_DIR, 'entities', 'simpleui_change_list_actions.html')


def create_dir_if_not_exist(dirpath: str):
    if not os.path.exists(dirpath):
        os.mkdir(dirpath)
        return False
    return True


create_dir_if_not_exist(TEMPDIR)


if __name__ == '__main__':
    run()


"""
依赖django的功能函数
"""

from . import pure

import math
import pandas as pd

from django.core.paginator import Paginator
from rest_framework.response import Response
from rest_framework.pagination import PageNumberPagination
from rest_framework.generics import GenericAPIView
from rest_framework.mixins import ListModelMixin, RetrieveModelMixin
from rest_framework.request import Request
from functools import wraps
from django.db import connection
from django.db.models import Q, F
from django.db.models import QuerySet
from django.db.models import Max
from django.conf import settings
from warnings import warn
from django.core.management.color import no_style
from django.db import connection

from rest_framework.renderers import JSONRenderer
from django.forms import model_to_dict
from django.db.models import QuerySet
from rest_framework.mixins import CreateModelMixin, UpdateModelMixin, DestroyModelMixin
from rest_framework import status
from django.db.models import Model
from rest_framework.exceptions import APIException
from rest_framework.exceptions import ErrorDetail
import re
from django.contrib.contenttypes.models import ContentType
import json
from rest_framework import serializers as s
from django.http.request import QueryDict
from .auth import get_my_api_error, my_api_assert_function
import re


def get_list(query_dc, key):
    if isinstance(query_dc, QueryDict):
        ret = query_dc.getlist(key)
    elif isinstance(query_dc, dict):
        ret = query_dc.get(key)
    else:
        raise TypeError('query_dc类型不明!')

    if isinstance(ret, str):
        ret = [ret]
    return ret


def get_base_serializer(base_model, base_fields='__all__'):
    """
    生成一个基础序列化器

    :param base_model: queryset或者base_model
    :param base_fields: 字段
    :return:
    """
    base_model = get_base_model(base_model)

    class BaseSerializer(s.ModelSerializer):
        class Meta:
            model = base_model
            fields = base_fields
    base_serializer = BaseSerializer
    return base_serializer


def judge_is_obj_level_of_request(request):
    """
    判断本次访问是否为obj对象级, 否则就是model模型级
    """
    if 'pk' in request._request.resolver_match.kwargs:
        return True
    else:
        return False


def conv_queryset_ls_to_serialzer_ls(qs_ls: list):
    """
    qs_ls: 多个queryset数据的序列化, 手动转化为dc_ls
    """
    dc_ls = []
    if not qs_ls:
        return dc_ls

    q = qs_ls[0]
    if not isinstance(q, dict):
        q = model_to_dict(q)
    kname_ls = list(q.keys())

    dc_ls = []
    for q in qs_ls:
        # print(q)
        for kname_i in kname_ls:
            dc = {
                kname_i: getattr(q, kname_i)
            }
            dc_ls.append(dc)
    return dc_ls


def get_field_type_in_db(model, field_name):
    """根据模型和字段名, 获取这个字段在数据库中对应的类型"""
    tp = model._meta.get_field(field_name).get_internal_type()
    return tp


def convert_db_field_type_to_python_type(tp):
    tp = re.sub(r'\(.*\)', '', tp)      # 删除括号内的内容, 如"CharField(source='more_group.explain') "
    if tp in ['TextField', 'CharField', 'DateField', 'FileField', 'DateTimeField']:
        field_type = 'str'
    elif tp in ['IntegerField', 'AutoField', 'BigAutoField']:
        field_type = 'int'
    elif tp in ['FloatField']:
        field_type = 'float'
    elif tp == 'BooleanField':
        field_type = 'bool'
    elif '=' in tp:
        # 类, 一般返回一个dc_ls类型
        field_type = 'list'
    else:
        field_type = tp
    return field_type


def get_field_type_in_py(model, field_name):
    """根据模型和字段名, 获取这个字段在python中对应的类型"""
    tp = get_field_type_in_db(model, field_name)
    field_type = convert_db_field_type_to_python_type(tp)
    return field_type


def reset_db_sequence(model):
    """重置数据库索引, 避免postgresql在手动导入csv/excel后出错."""
    md = get_base_model(model)
    sequence_sql = connection.ops.sequence_reset_sql(no_style(), [md])
    with connection.cursor() as cursor:
        for sql in sequence_sql:
            cursor.execute(sql)
    cursor.close()


def APIResponse(ret=None, status=200, msg=None):
    if isinstance(ret, Response):
        ret = ret.data
    ret = pure.add_status_and_msg(ret, status=status, msg=msg)
    ret = Response(ret)
    return ret


# --- 默认分页器参数设置 PAGINATION_SETTINGS
PAGINATION_SETTINGS = {
    'page_size': 16,       # 每页16个
    'page_size_query_param': 'page_size',       # 前端控制每页数量时使用的参数名, 'page_size'
    'page_query_param': 'p',        # 页码控制参数名"p"
    'max_page_size': 1000,      # 最大1000页
}

if hasattr(settings, 'PAGINATION_SETTINGS'):
    PAGINATION_SETTINGS.update(settings.PAGINATION_SETTINGS)


class Pagination(PageNumberPagination):
    """
    * 默认分页器参数设置

    - page_size: 每页16个
    - page_size_query_param: 前端控制每页数量时使用的参数名, 'page_size'
    - page_query_param: 页码控制参数名"p"
    - max_page_size: 最大1000页
    """
    page_size = int(PAGINATION_SETTINGS.get('page_size'))
    page_size_query_param = PAGINATION_SETTINGS.get('page_size_query_param', 'page_size')
    page_query_param = PAGINATION_SETTINGS.get('page_query_param', 'p')
    max_page_size = int(PAGINATION_SETTINGS.get('max_page_size'))


class StateMsgResultJSONRenderer(JSONRenderer):

    def render(self, data, accepted_media_type=None, renderer_context=None):
        if 'status' not in data and 'msg' not in data:
            if 'detail' in data:
                e = data.pop('detail')

                msg = str(e)
                if data:
                    msg += str(data)

                if isinstance(e, ErrorDetail) and e.code == 'permission_denied':
                    status = 403
                else:
                    print('! ************ 莫名返回格式 StateMsgResultJSONRenderer **************')
                    status = 404
                data = {
                    'status': status,
                    'msg': msg,
                    'result': [],
                }
            else:
                pass
        return super(StateMsgResultJSONRenderer, self).render(data, accepted_media_type, renderer_context)


class BaseListView(ListModelMixin, RetrieveModelMixin, GenericAPIView):
    """
    * API: BaseModel的ListView和RetrieveView接口

    - 测试接口:
        - List:
            GET /api/index/BaseList/?order_type=-id&page_size=4&p=1
        - Retrieve:
            GET /api/index/BaseList/5/
    """
    
    _name = 'BaseListView'      # 这个在自动生成wiki时要用到
    
    renderer_classes = (StateMsgResultJSONRenderer,)

    pagination_class = Pagination

    filter_fields = []       # ['__all__']过滤所有. 精确检索的过滤字段, 如果过滤条件复杂的话, 建议重写

    order_type_ls = []
    distinct_field_ls = []
    only_get_distinct_field = False     # 仅返回distinct指定的字段

    serializer_class = None
    auto_generate_serializer_class = False
    base_fields = '__all__'     # 当auto_generate_serializer_class为True时, 将自动生成序列化器, 然后根据base_fields返回字段

    retrieve_serializer_class = None       # retrieve对应的serializer_class
    list_serializer_class = None       # list对应的serializer_class

    retrieve_filter_field = None        # 详情页的查询字段名, 默认为 {{url}}/app/view/pk , 无需变动.

    method = None       # 中间变量, 记录请求是什么类型, list/retrieve等
    _tmp = None      # 无意义, 用来处理数据

    _post_type = None   # 用来判断是什么请求类型, 然后判断request_data取哪个
    post_type_ls = ["list", "retrieve", "bulk_list"]  # post请求方法
    create_unique = True  # 创建时是否允许重复

    def get(self, request: Request, *args, **kwargs):
        """
        - 如果request携带pk参数, 则进行Retrieve操作, 否则进行List操作.

        - BaseList的默认get请求参数(仅在List操作时生效)
            - page_size: 分页器每页数量, 前端用来控制数据每页展示的数量, 在Pagination类中设置.
            - p: 第p页.
            - order_type_ls: 排序字段, 如"id"和"-id".
        """
        if kwargs.get('pk'):
            self.method = 'retrieve'
            ret, status, msg = self.get_retrieve_ret(request, *args, **kwargs)
        else:
            self.method = 'list'
            ret, status, msg = self.get_list_ret(request, *args, **kwargs)
        return APIResponse(ret, status=status, msg=msg)

    def bulk_list(self, request, *args, **kwargs):
        """
        根据id批量删除
        """
        query_dc = self.get_request_data()

        id_ls = get_list(query_dc, 'id_ls')
        self.method = query_dc.get('http_method', 'retrieve')       # 优先返回retrieve详情数据

        my_api_assert_function(id_ls, msg=f'id_ls[{id_ls}]不能为空!!!')
        my_api_assert_function(isinstance(id_ls, list), msg=f'id_ls[{id_ls}]应为list类型, 不应为{id_ls.__class__.__name__}类型!!')

        base_model = get_base_model(self.queryset)
        qs_ls = base_model.objects.filter(id__in=id_ls)
        self.queryset = qs_ls

        qs_ls = self.get_ordered_queryset()

        page_size = query_dc.get('page_size', self.pagination_class.page_size)
        p = query_dc.get('p', 1)
        data, page_dc = paginate_qsls_to_dcls(qs_ls, self.get_serializer_class(), page=p, per_page=page_size)
        ret = {
            'page_dc': page_dc,
            'data': data
        }
        return APIResponse(ret)

    def post(self, request, *args, **kwargs):
        """用post方法来跳转"""
        post_type = request.data.get('post_type', 'list')
        self._post_type = post_type
        my_api_assert_function(not post_type or post_type in self.post_type_ls, f"操作类型post_type指定错误! 取值范围: {self.post_type_ls}")

        if post_type in ['list', 'retrieve']:
            return self.get(request, *args, **kwargs)
        elif post_type == 'bulk_list':
            return self.bulk_list(request, *args, **kwargs)
        else:
            return APIResponse(None, status=404, msg=f'请指定post操作类型, 取值范围: {self.post_type_ls}?')

    def get_serializer_class(self):
        if self.method == 'retrieve':
            ret = self.retrieve_serializer_class or self.serializer_class
        elif self.method == 'list':
            ret = self.list_serializer_class or self.serializer_class
        else:
            ret = self.retrieve_serializer_class or self.serializer_class or self.list_serializer_class

        if self.auto_generate_serializer_class and self.serializer_class is None:
            ret = get_base_serializer(self.queryset, base_fields=self.get_base_fields())

        # --- 仅获取distinct之后的字段
        only_get_distinct_field = self.get_only_get_distinct_field()
        if only_get_distinct_field:
            distinct_field_ls = self.get_distinct_field_ls()
            assert distinct_field_ls, '指定了only_get_distinct_field的同时必须指定distinct_field_ls!'

            if isinstance(distinct_field_ls, str):
                distinct_field_ls = [distinct_field_ls]

            ret = get_base_serializer(self.queryset, distinct_field_ls)
            return ret

        assert ret, '返回的serializer_class不能为空!'
        return ret

    def get_only_get_distinct_field(self):
        key = 'only_get_distinct_field'
        value = self.get_key_from_query_dc_or_self(key)
        ret = pure.convert_query_parameter_to_bool(value)
        return ret

    def get_distinct_field_ls(self):
        key = 'distinct_field_ls'
        ret = self.get_key_from_query_dc_or_self(key, get_type='list')
        return ret

    def get_base_fields(self):
        key = 'base_fields'
        ret = self.get_key_from_query_dc_or_self(key, get_type='list')
        return ret

    def get_retrieve_ret(self, request: Request, *args, **kwargs):
        """
        Retrieve操作

        - pk必须在`url.py::urlpatterns`中设置, 如: path('BaseList/<str:pk>/', views.BaseList.as_view())
        """
        status = 200
        msg = 'ok'

        try:
            ret = self.retrieve(request)
        except Exception as e:
            # 这里有坑, 因为有可能返回的是已经自定义好了的APIException, 例如权限不足错误.
            if not isinstance(e, APIException):
                # 没找到的情况, 404 Not Found
                ret = None
                status = 404
                msg = str(e)
                # msg = 'RetrieveError__' + str(e)
            else:
                raise e

        return ret, status, msg

    def get_object(self):
        """retrieve时, object的获取"""
        retrieve_filter_field = self.retrieve_filter_field if self.retrieve_filter_field else 'id'

        queryset = self.filter_queryset(self.get_queryset())
        pk = self.request.parser_context.get('kwargs').get('pk')
        queryset = get_base_model(queryset).objects.all()

        cmd = f'self._tmp=Q({retrieve_filter_field}={pk})'
        exec(cmd)

        queryset = queryset.filter(self._tmp)
        assert queryset.count() == 1, f'__{status.HTTP_404_NOT_FOUND}__: Error! 查询结果的数量应该等于1, 然而实际等于{queryset.count()}.'
        obj = queryset[0]
        self.check_object_permissions(self.request, obj)
        return obj

    def get_list_ret(self, request: Request, *args, **kwargs):
        """
        List操作
        """
        status = 200
        msg = 'ok'

        # --- 得到list方法的queryset
        self.queryset = self.get_list_queryset()

        # --- 根据ordering参数, 获得排序后的queryset
        self.queryset = self.get_ordered_queryset()

        # --- 获取返回数据(转化为特定格式)
        try:
            ret = self.list(request)
        except Exception as e:
            # 页码无效的情况, 404 Not Found
            ret = None
            status = 404
            msg = str(e)
        return ret, status, msg

    def get_request_data(self)->dict:
        """
        请求所携带的数据, 除了get方法跳过来的外, 均以请求体body携带的数据request.data优先.
        """
        if not hasattr(self, 'request'):
            return {}

        if self._post_type is None:
            ret = self.request.query_params
        else:
            # ret = self.request.GET or self.request.query_params or self.request.data
            ret = self.request.data or self.request.query_params

        if hasattr(self, '_set_request_data') and getattr(self, '_set_request_data'):
            ret = self._set_request_data
        return ret

    def set_request_data(self, request_data):
        self._set_request_data = request_data

    def get_ordered_queryset(self):
        """
        按order_type_ls指定的字段排序
        """
        query_dc = self.get_request_data()
        order_type_ls = get_list(query_dc, key='order_type_ls') if get_list(query_dc, key='order_type_ls') else self.order_type_ls
        distinct_field_ls = get_list(query_dc, key='distinct_field_ls') if get_list(query_dc, key='distinct_field_ls') else self.distinct_field_ls

        if not order_type_ls:
            # 旧版本可能用的order_type, 尝试赋值
            order_type = get_list(query_dc, key='order_type')
            if order_type:
                order_type_ls = order_type

        qs_ls = self.queryset

        # distinct操作
        if distinct_field_ls and distinct_field_ls not in ['__None__', ['__None__']]:
            assert isinstance(distinct_field_ls, (list, tuple)), 'distinct_field_ls因为list或者tuple!'
            qs_ls = qs_ls.order_by(*distinct_field_ls).distinct(*distinct_field_ls)      # bug: distinct_field_ls 后的字段无法排序

        # order_by操作
        if order_type_ls:
            if distinct_field_ls and distinct_field_ls not in ['__None__', ['__None__']]:
                qs_ls = self.queryset.filter(pk__in=qs_ls.values('pk'))

            try:
                qs_ls = order_by_order_type_ls(qs_ls, order_type_ls)
            except ValueError as e:
                msg = f'参数order_type_ls指定的排序字段[{order_type_ls}]排序失败! 更多信息: {str(e)}'
                raise ValueError(msg)

        self.queryset = qs_ls
        return self.queryset

    def get_list_queryset(self):
        """
        得到queryset, 仅对list方法生效
        """
        self.queryset = self.get_queryset()
        if not isinstance(self.queryset, QuerySet):
            self.queryset = self.queryset.objects.all()
        self.run_list_filter()
        return self.queryset

    def run_list_filter(self):
        """
        返回用self.filter_fields过滤后的queryset列表
        """
        if self.queryset.count():
            """
            过滤字段filter_fields
            """
            query_dc = self.get_request_data()
            FILTER_ALL_FIELDS = True if self.filter_fields in ['__all__', ['__all__']] else False

            self.queryset = self.get_queryset()
            base_model = get_base_model(self.queryset)
            if base_model.objects.count() == 0:
                return self.queryset

            meta = base_model.objects.first()._meta
            field_names = [field.name for field in meta.fields]
            field_names.append('pk')    # 将主键pk加进去作为过滤条件
            if self.queryset.count():
                qs_i = self.queryset[0]
                # 可能用annotate增加了注释字段, 所以要处理一下, 避免过滤出错
                if hasattr(qs_i, 'keys') and len(qs_i.keys()) > len(field_names):
                    field_names = list(qs_i.keys())

            for fn in field_names:      # fn: field_name
                if FILTER_ALL_FIELDS or fn in self.filter_fields or fn == 'pk':
                    if fn in query_dc:
                        value = query_dc.get(fn)
                        if value is not None and value != '':       # 默认为空字符串时, 将不作为过滤条件
                            dc = {fn: value}
                            q = Q(**dc)
                            self.queryset = self.queryset.filter(q)
                            # ss = f"self.queryset = self.queryset.filter(Q({fn}='{value}'))"
                            # exec(f"self.queryset = self.queryset.filter(Q({fn}='{value}'))")
        return self.queryset

    def _get_list_queryset(self):       # 兼容问题
        return self.get_list_queryset()

    def list(self, request: Request, *args, **kwargs):
        query_dc = self.get_request_data()
        page_size = query_dc.get('page_size', self.pagination_class.page_size)
        p = query_dc.get('p', 1)
        context = self.get_serializer_context()
        resp, page_dc = paginate_qsls_to_dcls(self.queryset, self.get_serializer_class(), page=p, per_page=page_size, context=context)
        ret = self._conv_data_format(resp, page_dc)
        return ret

    def conv_data_format(self, data: (dict, Response), page_dc):
        ret = {
            'page_dc': page_dc,
            'data': data,
        }
        return ret

    def _conv_data_format(self, *args, **kwargs):
        return self.conv_data_format(*args, **kwargs)

    def get_key_from_query_dc_or_self(self, key, get_type=None):
        """
        优先检索query_dc是否有key, 其次检索self是否有key这个属性
        :param key: 变量名
        :return:
        """

        query_dc = self.get_request_data()

        if get_type == 'list':
            data = get_list(query_dc, key)
        elif get_type == 'bool':
            data = pure.convert_query_parameter_to_bool(query_dc.get(key))
        else:
            data = query_dc.get(key)

        # 让request携带的数据可以覆盖自身的key值
        ret_0 = getattr(self, key) if hasattr(self, key) else None
        ret_1 = data if data else None

        ret = ret_1 or ret_0
        return ret

    def _get_key_from_query_dc_or_self(self, *args, **kwargs):
        return self.get_key_from_query_dc_or_self(*args, **kwargs)


class BaseList(BaseListView):     # 向下兼容, 返回格式调整, 重写_conv_data_format.
    def _conv_data_format(self, data: (dict, Response)):
        if isinstance(data, Response):
            data = data.data

            # 分页信息
        count = data.get('count')
        page_size = self.request.query_params.get('page_size', self.pagination_class.page_size)
        p = self.request.query_params.get('p', 1)
        total = math.ceil(count / int(page_size))
        page_dc = {
            'count': count,
            'total': total,
            'page_size': page_size,
            'p': p,
        }

        results = data.get('results')

        ret = {
            'page_dc': page_dc,
            'results': results,
        }
        return ret


def get_base_model(obj) -> Model:
    """判断是Queryset还是BaseModel"""
    if isinstance(obj, QuerySet):
        return obj.model
    else:
        if isinstance(obj, ContentType):
            base_model = obj.model_class()
            return base_model

        if hasattr(obj, 'objects'):
            # BaseModel
            return obj
        elif hasattr(obj.__class__, 'objects'):
            # 单个obj
            return obj.__class__
        else:
            return obj


def get_base_queryset(obj) -> QuerySet:
    """
    返回所有obj类型的QuerySet
    """
    ret = get_base_model(obj)
    ret = ret.objects.all()
    return ret


def paginate_qsls_to_dcls(qsls, serializer, page: int, per_page=16, context=None):
    """
    * 手动分页函数

    - 指定模型的queryset_ls和serializer, 然后按给定的page和per_page参数获取分页后的数据
    """

    p = Paginator(qsls, per_page)
    page_obj = p.get_page(page)
    # page_dc = {
    #     'num_pages': p.num_pages,
    #     'count_objects': p.count,
    #     'current_page_number': page_obj.number,
    # }
    page_dc = {
            "count_items": int(p.count),
            "total_pages": int(p.num_pages),
            "page_size": int(per_page),
            "p": int(page)
        },

    # --- 处理单个Model和多个Model的情况
    if serializer.__class__.__name__ == 'function':
        try:
            dc_ls = serializer(page_obj, context=context)
        except Exception as e:
            print('--- paginate_qsls_to_dcls错误!!! 2022/2/25')
    else:
        dc_ls = serializer(page_obj, many=True, context=context).data
    return dc_ls, page_dc


def conv_queryset_to_dc_ls(queryset: QuerySet):
    dc_ls = []
    for q in queryset:
        dc_ls.append(q)
    return dc_ls


def order_qs_ls_by_id(qs_ls, sort_by='id'):
    df = pd.DataFrame(qs_ls).sort_values(by=sort_by)

    cols = df.columns
    dc_ls = []
    for i, row in df.iterrows():
        dc = {
            cols[0]: row.get(cols[0]),
            cols[1]: row.get(cols[1]),
        }
        dc_ls.append(dc)
    return dc_ls


def order_by_order_type_ls(queryset, order_type_ls) -> QuerySet:
    """
    根据传入的order_type_ls参数, 对queryset进行排序.

    若order_type_ls包含__None__, 则清空queryset的排序规则.
    """
    if '__None__' in order_type_ls:
        ret = queryset.order_by()
    elif order_type_ls is None or isinstance(order_type_ls, str):
        ret = order_by_order_type(queryset, order_type_ls)
    else:
        ls = []
        for order_type in order_type_ls:
            if order_type:
                if order_type.startswith('-'):
                    order_type1 = order_type[1:]
                    ls.append(F(order_type1).desc(nulls_last=True))
                else:
                    ls.append(F(order_type).asc(nulls_first=True))
        ret = queryset.order_by(*ls)
    return ret


def order_by_order_type(queryset, order_type=None):
    ret = queryset
    if order_type:
        if order_type.startswith('-'):
            order_type1 = order_type[1:]
            ret = queryset.order_by(F(order_type1).desc(nulls_last=True))
        else:
            ret = queryset.order_by(F(order_type).asc(nulls_first=True))

    return ret


def api_decorator(func):
    """
    * API装饰器

    - 如果运行出错, 将格式化输出错误的信息, 并返回给前端, 而不会报错.
    - 自动处理postgresql中idle状态connection过多的情况
    """
    @wraps(func)
    def wrapped_function(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            print('--- API Error! ---')
            print(e)
            msg = f'Error! {str(e)}'

            e_str = str(e)
            if 'client' in e_str:
                msg += '!!!!!! 可能出现postgresql的idle链接状况???'
                print(msg)
                # --- postgres的idle链接需要解决, 关闭旧链接(以下使用), 或单线程运行`manage.py runserver --nothreading`
                from django.db import close_old_connections
                from django.db import connection
                close_old_connections()
                with connection.cursor() as cursor:
                    sql = "SELECT pg_terminate_backend(pid) FROM pg_stat_activity WHERE state = 'idle'"
                    cursor.execute(sql)
                    row = cursor.fetchall()
                    print(sql)
                    print(row)
            my_api_assert_function(False, msg=msg, status='404')
    return wrapped_function


def get_model_max_id_in_db(model):
    """
    仅适用于postgresql, mysql直接忽略就行.
    """
    meta = model._meta
    if not callable(model):
        model = type(model)
    ordering = meta.ordering

    assert sum([1 if f.name == 'id' and f.primary_key is True else 0 for f in meta.fields]), "Model的主键必须是id!"

    if ordering:
        if ordering[0] == '-id':
            obj = model.objects.first()
            max_id = obj.id if obj else 0
            ret = max_id + 1
            return ret
        if ordering[0] == 'id':
            obj = model.objects.last()
            max_id = obj.id if obj else 0
            ret = max_id + 1
            return ret
    qs = model.objects.all()
    max_id = qs.aggregate(max_id=Max('id')).get('max_id')  if qs.count() else 0
    ret = max_id + 1
    return ret


def old_get_model_max_id_in_db(model):
    meta = model._meta

    assert sum([1 if f.name == 'id' and f.primary_key is True else 0 for f in meta.fields]), "Model的主键必须是id!"

    cursor = connection.cursor()
    # db_prefix = meta.__str__().split('.')[0]

    # --- 先尝试创建id_seq
    id_seq = f"{meta.app_label}_{meta.db_table}_id_seq"

    try:
        sql = f"""CREATE SEQUENCE IF NOT EXISTS {id_seq}"""
        cursor.execute(sql)
    except Exception as e:
        # mysql不执行本函数也能正常运行
        print(e)
        print('不是PostgreSQL无法运行CREATE SEQUENCE语句! 请确认数据库类型!')

    # --- 找出最大的id
    sql = f"""select setval('{id_seq}', (select max(id) from "{meta.db_table}")+1);"""
    print('---', sql)
    # sql = '(select max(id) from  "{meta.db_table}")'
    # print('sql---', sql)
    cursor.execute(sql)
    row = cursor.fetchall()
    curr_id = row[0][0]
    ret = 0 if curr_id is None else curr_id
    cursor.close()
    return ret


def get_abs_order_type_ls(order_type_ls):
    if isinstance(order_type_ls, str):
        order_type_ls = [order_type_ls]
    ret = [re.sub(r'^-', '', field_name) for field_name in order_type_ls]
    return ret


from .mixins import MyCreateModelMixin, MyUpdateModelMixin, MyDestroyModelMixin


class CompleteModelView(BaseListView, MyCreateModelMixin, MyUpdateModelMixin, MyDestroyModelMixin):
    """
    * 一个模型的增删改查全套接口

    - 可在post方法中使用post_type指定操作类型.

    > 以下方法分别对应同一个url下的: 增, 删, 改, 查.

    - POST
        - 创建新数据
    - DELETE
        - 删除数据, 需指定id
    - PUT
        - 修改数据, 需指定id
    - GET
        - 查询列表页`GET url/`
        - 查询详情页, 需指定id.
          - 如: `GET url/id/`
    """
    _name = 'CompleteModelView'

    post_type_ls = ["list", "retrieve", "create", "update", "delete", "bulk_delete", "bulk_update", "bulk_list"]       # post请求方法
    _post_type = None
    create_unique = True        # 创建时是否允许重复

    def post(self, request, *args, **kwargs):
        """增"""
        post_type = request.data.get('post_type', 'create')
        self._post_type = post_type
        my_api_assert_function(not post_type or post_type in self.post_type_ls, f"操作类型post_type指定错误! 取值范围: {self.post_type_ls}")

        if post_type == 'create':
            return self.create(request, *args, **kwargs)
        elif post_type == 'delete':
            return self.destroy(request, *args, **kwargs)
        elif post_type in ['list', 'retrieve']:
            return self.get(request, *args, **kwargs)
        elif post_type == 'update':
            return self.put(request, *args, **kwargs)
        elif post_type == 'bulk_delete':
            return self.bulk_delete(request, *args, **kwargs)
        elif post_type == 'bulk_update':
            return self.bulk_update(request, *args, **kwargs)
        elif post_type == 'bulk_list':
            return self.bulk_list(request, *args, **kwargs)
        else:
            return APIResponse(None, status=404, msg='请指定post操作类型, [create, update, delete]?')

    def delete(self, request, *args, **kwargs):
        """删"""
        return self.destroy(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        """改"""
        ret = self.partial_update(request, *args, **kwargs)
        return ret

    def bulk_delete(self, request, *args, **kwargs):
        """
        根据id批量删除
        """
        request_data = self.get_request_data()

        id_ls = get_list(request_data, 'id_ls')
        my_api_assert_function(id_ls, msg=f'id_ls[{id_ls}]不能为空!!!')
        my_api_assert_function(isinstance(id_ls, list), msg=f'id_ls[{id_ls}]应为list类型, 不应为{id_ls.__class__.__name__}类型!!')

        base_model = get_base_model(self.queryset)
        qs_ls = base_model.objects.filter(id__in=id_ls)
        qs_ls.delete()
        return APIResponse()

    def bulk_update(self, request, *args, **kwargs):
        """
        批量更新
        """
        request_data = self.get_request_data()

        id_ls = get_list(request_data, 'id_ls')
        field_dc = request_data.get('field_dc')

        my_api_assert_function(id_ls, msg=f'id_ls[{id_ls}]不能为空!!!')
        my_api_assert_function(isinstance(id_ls, list), msg=f'id_ls[{id_ls}]应为list类型, 不应为{id_ls.__class__.__name__}类型!!')
        my_api_assert_function(field_dc, 'field_dc不能为空!')

        if isinstance(field_dc, str):
            field_dc = json.loads(field_dc)

        # 开始批量更新, 注意只支持2层嵌套.
        foreign_key_field_dc = {}
        original_field_dc = {}
        for k, v in field_dc.items():
            dc = {k: v}
            if '__' in k:
                foreign_key_field_dc.update(dc)
            else:
                original_field_dc.update(dc)

        base_model = get_base_model(self.queryset)
        qs_ls: m.QuerySet = base_model.objects.filter(id__in=id_ls)
        my_api_assert_function(qs_ls.count(), '未找到id_ls对应的数据!')

        qs_i = qs_ls[0]     # 样例数据

        # 先更新原生字段
        qs_ls.update(**original_field_dc)

        # 再依次更新外键字段
        for k, v in foreign_key_field_dc.items():
            fk_model_name, fk_field = k.split('__')
            dc = {fk_field: v}
            foreign_key_obj_i = getattr(qs_i, fk_model_name)
            foreign_key_model = get_base_model(foreign_key_obj_i)
            base_field = getattr(base_model, fk_model_name)
            assert hasattr(base_field, 'related'), '字段不是外键?'
            related = getattr(base_field, 'related')
            foreign_key_field_name = related.field.name
            filter_dc = {
                f'{foreign_key_field_name}__id__in': id_ls
            }
            qs_ls = foreign_key_model.objects.filter(**filter_dc)
            qs_ls.update(**dc)

        # 返回更新后的数据
        qs_ls: m.QuerySet = base_model.objects.filter(id__in=id_ls)
        ret = self.get_serializer_class()(qs_ls, many=True).data
        return APIResponse(ret)


class DecoratorBaseListView(BaseListView):
    @api_decorator
    def get(self, request, *args, **kwargs):
        return super().get(*args, **kwargs)

    @api_decorator
    def post(self, request, *args, **kwargs):
        return super().post(*args, **kwargs)


class DecoratorCompleteModelView(CompleteModelView):
    """
    全部用api_decorator装饰
    """

    @api_decorator
    def get(self, request, *args, **kwargs):
        return super().get(*args, **kwargs)

    @api_decorator
    def post(self, *args, **kwargs):
        return super().post(*args, **kwargs)

    @api_decorator
    def put(self, request, *args, **kwargs):
        return super().put(*args, **kwargs)

    @api_decorator
    def delete(self, request, *args, **kwargs):
        return super().delete(*args, **kwargs)


from django.db import connection


def get_executable_sql(queryset):
    """
    输入queryset, 获得可直接执行的sql语句
    """
    cursor = connection.cursor()
    sql, params = queryset.query.sql_with_params()
    prefix = 'EXPLAIN '
    cursor.execute(prefix + sql, params)
    sql: str = cursor.db.ops.last_executed_query(cursor, sql, params)
    sql = sql[len(prefix):]
    cursor.close()
    return sql


from django.db import models as m


def get_MySubQuery(my_model, field_name, function_name, output_field=m.IntegerField, alias=None):
    """
    # 获取子查询

    ## 简介
    - 主要用在进行`qs_ls.all().order_by().order_by(field_name).distinct(field_name)`后, 再进行`annotate`操作.
    - 普通的`m.SubQuery`操作将在`distinct`后的`annotate`中报错!

    ## 参考
    - [django文档_1](https://django-orm-cookbook-zh-cn.readthedocs.io/zh_CN/latest/subquery.html)
    - [django文档_2](https://docs.djangoproject.com/zh-hans/4.0/ref/models/expressions/)

    :param my_model: 指定模型, 用以获取模型基本属性`meta`. 若为空, 则认为是annotate字段, 使用默认的field_name.
    :param field_name: 用来进行计算的字段名
    :param function_name: 要在数据库中调用的函数名
    :param output_field: 使用Query计算后, 输出的字段类型
    :param alias: 计算后储存结果变量名, 默认为`tmp`
    :return: 子查询类`MySubQuery`
    """
    base_model = get_base_model(my_model)
    meta = base_model._meta

    field_names = [field.name for field in meta.fields]
    db_column_names = [field.db_column if field.db_column else field.name for field in meta.fields]
    field_dc = dict(zip(field_names, db_column_names))
    db_column_name = field_dc.get(field_name)  # 获取字段在db中的列名
    db_column_name = db_column_name if db_column_name else field_name       # 没有的话, 就用默认field_name

    alias = 'tmp' if not alias else alias
    my_template = f"(SELECT {function_name}({db_column_name}) FROM (%(subquery)s) {alias})"
    my_output_field = output_field

    class MySubQuery(m.Subquery):
        template = my_template
        output_field = my_output_field() if isinstance(my_output_field, type) else my_output_field

    return MySubQuery


def get_obj_by_content_type(obj_id, model_name, app_label):
    ct_qs_ls = ContentType.objects.filter(app_label=app_label, model=model_name)
    assert ct_qs_ls.count() == 1, f'ContentType数量不为1! current_value: {ct_qs_ls.count()}'
    ct_qs_i = ct_qs_ls[0]
    base_model = ct_qs_i.model_class()
    obj = base_model.objects.get(id=obj_id)
    return obj


def get_QS_by_dc(dc, add_type):
    """
    根据dc返回QS
    :param dc: 过滤条件
    :param add_type: 合并逻辑
    :return: QS
    """
    QS = Q()
    for k, v in dc.items():
        d = {k: v}
        QS.add(Q(**d), add_type)
    return QS


def get_model_verbose_name_dc():
    """
    获得model_verbose_name对应的ContentType的id
    """
    ct_qs_ls = ContentType.objects.all()
    dc = {}
    for ct_qs_i in ct_qs_ls:
        base_model = ct_qs_i.model_class()
        if base_model is not None:
            k = base_model._meta.verbose_name
            # v = ct_qs_i.model
            v = ct_qs_i
            dc.update({k: v})
    return dc


def get_user_ip(request):
    context = request.parser_context
    if 'HTTP_X_FORWARDED_FOR' in context["request"].META:
        user_ip = context["request"].META['HTTP_X_FORWARDED_FOR']
    else:
        user_ip = context["request"].META['REMOTE_ADDR']
    return user_ip


def update_none_to_zero_by_field_name(qs_ls, field_name):
    """
    将qs_ls中field_name字段的None改为0
    """
    filter_dc = {
        f'{field_name}__isnull': True
    }
    update_dc = {
        field_name: 0
    }
    qs_ls.filter(**filter_dc).update(**update_dc)


def get_df_by_freq_and_year(
        queryset,
        frequency_cname=None,
        aggregate_method_name='Sum',
        output_col_name=None,
        year_field_name='year',
        complete_year_ls=False,
        year_range_ls=None,
):
    """
    获取加权的年份分布图

    * 这个比较快

    - aggregate_method_name, frequency_cname: 用来aggregate的方法和字段
    - complete_year_ls: 补全中间年份
    - output_col_name: 输出列名
    - year_field_name: 年份字段名
    """
    year_qsv_ls = queryset.values(year_field_name).distinct(year_field_name).order_by(year_field_name)
    year_ls = [dc.get(year_field_name) for dc in year_qsv_ls]
    assert hasattr(m, aggregate_method_name), f'django.db.models不存在[{aggregate_method_name}]方法!'
    aggregate_method = getattr(m, aggregate_method_name)

    output_col_name = output_col_name if output_col_name  else frequency_cname

    if complete_year_ls:
        if year_range_ls:
            year_min, year_max = year_range_ls
        else:
            year_min, year_max = min(year_ls), max(year_ls)
        year_range = list(range(year_min, year_max + 1))
    else:
        year_range = year_ls

    year_distribution_dc_ls = []
    for year in year_range:
        aggregate_dc = {'tmp': aggregate_method(frequency_cname)}
        value = queryset.filter(year=year).aggregate(**aggregate_dc)
        value = 0 if value is None or value.get('tmp') is None else value.get('tmp')

        year_distribution_dc = {
            year_field_name: year,
            output_col_name: value,
        }
        year_distribution_dc_ls.append(year_distribution_dc)
    return year_distribution_dc_ls



import os
import setuptools
import bddjango
from m2r import parse_from_file


dirname = 'bddjango'
version = bddjango.version()
# print("version ================ ", version)


# with open("Readme.md", "r", encoding='utf-8') as fh:
#     long_description = fh.read()
long_description = parse_from_file('Readme.md')     # .markdown必须转换为.rst, 否则有可能报错


setuptools.setup(
    name=dirname,
    version=version,
    author="bode135",
    author_email='2248270222@qq.com',   # 作者邮箱
    description="常用的django开发工具",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://gitee.com/bode135/bdtools/tree/master/bddjango',   # 主页链接
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=['pandas', 'django', 'openpyxl', 'xlrd==1.2.0', 'bdtime', 'tqdm'],      # 依赖模块
    include_package_data=True,
)
